#ifndef _CLM_H_
#define _CLM_H_

#define CL_TARGET_OPENCL_VERSION 200
#include <CL/cl_platform.h>
#include <stdbool.h>

#define matrixAt(mat, r, c) mat.values[(!mat.transposed ? r * mat.cols + c : c * mat.rows + r)]

typedef struct clm_NativeBuf clm_NativeBuf;

typedef struct {
	float *values;
	unsigned int rows;
	unsigned int cols;
	bool transposed;
} clm_Matrix;

typedef struct {
	float *values;
	unsigned int length;
} clm_Vector;

typedef struct {
	clm_Matrix weights;
	clm_Matrix bias;
	clm_Matrix *output;
	clm_Matrix *error;
	clm_Matrix *weightsError;
	clm_Matrix *gradient;

	clm_NativeBuf *nativeWeights;
	clm_NativeBuf *nativeBias;
	clm_NativeBuf *nativeInput;
	clm_NativeBuf *nativeOutput;
	clm_NativeBuf *nativeInputError;
	clm_NativeBuf *nativeOutputError;
	clm_NativeBuf *nativeWeightsError;
	clm_NativeBuf *nativeGradient;
} clm_Linear;

typedef struct {
	clm_Linear *layers;
	unsigned int numLayers;
	float learnRate;
	unsigned int batchSize;
} clm_NN;

extern const clm_Matrix INVALID_MATRIX;
extern const clm_Vector INVALID_VECTOR;

clm_Matrix clm_createMatrixRandom(unsigned int rows, unsigned int cols);
clm_Matrix clm_matrixZero(clm_Matrix mat);
clm_Matrix clm_matrixAddScalar(clm_Matrix mat, float scalar);
clm_Matrix clm_matrixAddMatrix(clm_Matrix mat, clm_Matrix other);
clm_Matrix clm_matrixSubtractMatrix(clm_Matrix mat, clm_Matrix other);
clm_Matrix clm_matrixTranspose(clm_Matrix mat);
clm_Matrix clm_matrixMultiplyMatrix(clm_Matrix a, clm_Matrix b, clm_Matrix out);
clm_Matrix clm_matrixMultiplyMatrixALLOC(clm_Matrix a, clm_Matrix b);
clm_Matrix clm_matrixMultiplyMatrixElements(clm_Matrix mat, clm_Matrix other);
clm_Matrix clm_matrixMultiplyScalar(clm_Matrix mat, float scalar);
clm_Matrix clm_matrixSigmoid(clm_Matrix mat);
clm_Matrix clm_matrixDSigmoid(clm_Matrix mat);

clm_Matrix clm_matrixFromArray(float *array, unsigned int length);
clm_Matrix clm_matrixWrapArray(float *array, unsigned int length);
clm_Matrix clm_matrixCopyALLOC(clm_Matrix matrix);
clm_Matrix clm_matrixCopy(clm_Matrix mat, clm_Matrix other);

bool clm_matrixIsInvalid(clm_Matrix mat);

clm_Vector clm_vectorCreate(unsigned int length);

bool clm_vectorIsInvalid(clm_Vector vec);

clm_Linear clm_linearCreateRandom(unsigned int inputs, unsigned int outputs);
clm_NN clm_nnCreate(unsigned int numLayers, clm_Linear *layers, float learnRate, unsigned int batchSize);

void clm_matrixPrint(clm_Matrix mat);
void clm_matrixPrintShape(clm_Matrix mat);
void clm_freeMatrix(clm_Matrix mat);
void clm_freeVector(clm_Vector vector);
void clm_freeLinear(clm_Linear linear);

#endif
